@@ -22,10 +22,12 @@ typedef struct {
 
 typedef struct {
     SV *(*call_p6_method)(IV, char *, I32, SV *, SV **);
+    SV *(*call_p6_package_method)(char *, char *, I32, SV *, SV **);
     SV *(*call_p6_callable)(IV, SV *, SV **);
     void (*free_p6_object)(IV);
     SV *(*hash_at_key)(IV, char *);
     SV *(*hash_assign_key)(IV, char *, SV *);
+    SV *(*compile_to_end)(char *, char *, U32 *pos);
 } perl6_callbacks;
 
 XS(p5_call_p6_method);
@@ -69,6 +71,33 @@ size_t p5_size_of_nv() {
 
 static int inited = 0;
 
+static Perl_keyword_plugin_t next_keyword_plugin;
+static int raku_keyword_plugin(pTHX_ char *keyword_ptr, STRLEN keyword_len, OP **op_ptr) {
+    if (memEQs(keyword_ptr, keyword_len, "raku")) {
+        // read the whole source file
+        while (lex_next_chunk(0));
+
+        // move into the raku block
+        lex_read_space(0);
+        lex_read_to(PL_parser->bufptr + 1);
+
+        declare_cbs;
+        U32 pos;
+        char *bufptr = PL_parser->bufptr;
+        STRLEN len;
+        char *package_name = SvPV(PL_curstname, len);
+        SV *code = cbs->compile_to_end(package_name, PL_parser->bufptr, &pos);
+        lex_read_to(bufptr + pos + 1);
+
+        *op_ptr = newUNOP(OP_ENTERSUB, 0, newSVOP(OP_CONST, 0, code));
+
+        return KEYWORD_PLUGIN_STMT;
+    }
+    else {
+        return next_keyword_plugin(aTHX_ keyword_ptr, keyword_len, op_ptr);
+    }
+}
+
 void p5_inline_perl6_xs_init(PerlInterpreter *my_perl) {
     char *file = __FILE__;
     newXS("Perl6::Object::call_method", p5_call_p6_method, file);
@@ -83,17 +112,21 @@ void p5_inline_perl6_xs_init(PerlInterpreter *my_perl) {
 
 void p5_init_callbacks(
     SV  *(*call_p6_method)(IV, char * , I32, SV *, SV **),
+    SV  *(*call_p6_package_method)(char *, char * , I32, SV *, SV **),
     SV  *(*call_p6_callable)(IV, SV *, SV **),
     void (*free_p6_object)(IV),
     SV  *(*hash_at_key)(IV, char *),
-    SV  *(*hash_assign_key)(IV, char *, SV *)
+    SV  *(*hash_assign_key)(IV, char *, SV *),
+    SV  *(*compile_to_end)(char *, char *, U32 *)
 ) {
     perl6_callbacks *cbs = malloc(sizeof(perl6_callbacks));
     cbs->call_p6_method   = call_p6_method;
+    cbs->call_p6_package_method = call_p6_package_method;
     cbs->call_p6_callable = call_p6_callable;
     cbs->free_p6_object   = free_p6_object;
     cbs->hash_at_key      = hash_at_key;
     cbs->hash_assign_key  = hash_assign_key;
+    cbs->compile_to_end   = compile_to_end;
     hv_stores(PL_modglobal, "Inline::Perl5 callbacks", newSViv((IV)cbs));
 }
 
@@ -104,10 +137,12 @@ PerlInterpreter *p5_init_perl(
     int argc,
     char **argv,
     SV  *(*call_p6_method)(IV, char * , I32, SV *, SV **),
+    SV  *(*call_p6_package_method)(char *, char * , I32, SV *, SV **),
     SV  *(*call_p6_callable)(IV, SV *, SV **),
     void (*free_p6_object)(IV),
     SV  *(*hash_at_key)(IV, char *),
-    SV  *(*hash_assign_key)(IV, char *, SV *)
+    SV  *(*hash_assign_key)(IV, char *, SV *),
+    SV  *(*compile_to_end)(char *, char *, U32 *)
 ) {
     if (inited) {
 #ifndef MULTIPLICITY
@@ -131,12 +166,16 @@ PerlInterpreter *p5_init_perl(
 
     p5_init_callbacks(
         call_p6_method,
+        call_p6_package_method,
         call_p6_callable,
         free_p6_object,
         hash_at_key,
-        hash_assign_key
+        hash_assign_key,
+        compile_to_end
     );
 
+    wrap_keyword_plugin(raku_keyword_plugin, &next_keyword_plugin);
+
     return my_perl;
 }
 
@@ -464,6 +503,15 @@ const char *p5_sv_reftype(PerlInterpreter *my_perl, SV *sv) {
     return sv_reftype(SvRV(sv), 1);
 }
 
+SV *p5_new_blessed_hashref(PerlInterpreter *my_perl, char *package) {
+    PERL_SET_CONTEXT(my_perl);
+    HV * obj = newHV();
+    SV * const inst = newRV_noinc((SV*)obj);
+    HV *stash = gv_stashpv(package, GV_ADD);
+    (void)sv_bless(inst, stash);
+    return inst;
+}
+
 I32 p5_get_type(PerlInterpreter *my_perl, SV *sv) {
     int is_hash;
     PERL_SET_CONTEXT(my_perl);
@@ -538,6 +586,74 @@ void p5_set_global(PerlInterpreter *my_perl, const char* name, SV *value) {
         croak("Setting global hash variable NYI");
 }
 
+I32 p5_compile_sv(PerlInterpreter *my_perl, SV *line, CV **cv, SV **stash) {
+    PERL_SET_CONTEXT(my_perl);
+    char *start = NULL;
+    char start_orig;
+    I32 floor;
+    OP *op;
+
+    ENTER;
+
+    floor = start_subparse(FALSE, 0);
+
+    if (PL_parser) {
+        start = strstr(PL_parser->bufptr, SvPVX(line) + 1);
+        start_orig = start[-1];
+        start[-1] = '{';
+        lex_read_to(start - 1);
+    }
+    else {
+        lex_start(line, NULL, 0);
+    }
+
+    HV *new_stash = newHV();
+    HV *cur_stash = PL_curstash;
+    PL_curstash = new_stash;
+
+    op = parse_block(0);
+
+    PL_curstash = cur_stash;
+    HE *he;
+    while ((he = hv_iternext(new_stash))) {
+        SV *key = hv_iterkeysv(he);
+        SV *val = hv_iterval(new_stash, he);
+
+        /* Put the entry into the actual target stash */
+        hv_store_ent(PL_curstash, key, SvREFCNT_inc(val), HeHASH(he));
+
+        if (   0 == strcmp(SvPV_nolen(key), "BEGIN")
+            || 0 == strcmp(SvPV_nolen(key), "END")
+            || 0 == strcmp(SvPV_nolen(key), "__ANON__"))
+            hv_delete_ent(new_stash, key, G_DISCARD, HeHASH(he));
+    }
+    *stash = sv_2mortal(newRV_noinc((SV*)new_stash));
+
+    I32 remainder = PL_parser->bufend - PL_parser->bufptr;
+
+    *cv = newATTRSUB(floor, NULL, NULL, NULL, op);
+
+    if (start)
+        start[-1] = start_orig;
+    else
+        remainder -= 2; /* not sure why but it fixes things */
+
+    LEAVE;
+
+    return remainder;
+}
+
+void p5_runops(PerlInterpreter *my_perl, CV *cv) {
+    dSP;
+    ENTER;
+    SAVETMPS;
+    PUSHMARK(sp);
+    call_sv((SV*)cv, G_DISCARD);
+    SPAGAIN;
+    FREETMPS;
+    LEAVE;
+}
+
 SV *p5_eval_pv(PerlInterpreter *my_perl, const char* p, I32 croak_on_error) {
     PERL_SET_CONTEXT(my_perl);
     {
@@ -1563,15 +1679,26 @@ XS(p5_call_p6_extension_method) {
     SV * static_class = ST(1);
     SV * name = ST(2);
     SV * obj = ST(3);
+    SV * err = NULL;
 
-    if (!SvROK(obj)) {
-        croak("Got a non-reference for obj in p5_call_p6_extension_method?!");
-    }
-    MAGIC * mg = find_shadow_magic(p6cb, static_class, obj);
     STRLEN len;
     char * const name_pv  = SvPV(name, len);
+
+    if (!SvROK(obj)) {
+        if (SvPOK(obj)) {
+            char * const package_pv  = SvPV(obj, len);
+            AV *args = create_args_array(ax, items, 4);
+            SV * const args_rv = newRV_noinc((SV *) args);
+            declare_cbs;
+            SV * retval = cbs->call_p6_package_method(package_pv, name_pv, GIMME_V == G_SCALAR, args_rv, &err);
+            return post_callback(ax, sp, items, args_rv, err, retval);
+        }
+        else {
+            croak("Got a non-reference for obj in p5_call_p6_extension_method?!");
+        }
+    }
+    MAGIC * mg = find_shadow_magic(p6cb, static_class, obj);
     _perl6_magic* const p6mg = (_perl6_magic*)(mg->mg_ptr);
-    SV *err = NULL;
 
     AV *args = create_args_array(ax, items, 4);
     SV * const args_rv = newRV_noinc((SV *) args);
